function [ position_max, position_min, g ] = ...
  constantLinearMIE(linear_constant, input_bounds, dynamics_term)
% constantLinearMIE: double integrator reachability (1D MIE scheme)
%
%   [ position_max, position_min, g ] = ...
%     constantLinearMIE(linear_constant, input_bounds, dynamics_term)
%
% This is a reachability calculation using the Mixed Implicit Explicit
% scheme for the constant linear example, which is essentially the double
% integrator with an additional constant linear term in the dynamics for
% the integrating variable (the terms "position" and "velocity" don't make
% quite as much sense anymore).
%
%     \dot x_1 = a * x_1 + x_2 + v  (the horizontal direction; 'position')
%     \dot x_2 = u                  (the vertical direction; 'velocity')
%
%       with a given by linear_constant 
%       and bounds on the inputs v and u are given by input_bounds
%
% The x_1 variable is the "linear" variable which is treated specially -- its
% maximum (or minimum) value is tracked as the solution to a HJ PDE in the
% "nonlinear" x_2 variables.  In this particular case, x_2 happens to have
% a constant (zero) derivative with respect to the other variables, but we
% do not take advantage of that fact.
%
% Note that position_max and position_min are defined everywhere in the
% computational domain; however, only position values that lie in the
% interval [ position_min, position_max ] are within the reach set.
% Notably, if position_min > position_max for some points in the
% computational domain, then the reach set is empty at those points.
%
% The default settings (1 & [ 0.5, 0.25 ]) are chosen to correspond to
% an example which is supported by HJ PDE theory, but has unstable forward
% linear dynamics.  An example which is not supported by HJ PDE theory but
% is forward stable is (-1 & [ 0.0, 0.1 ]).  In both cases the reachable
% sets are noticably different than the regular double integrator and each
% other, but they are non-trivial (eg: not empty or the same as the target
% set).
%
% There are many options in the function which can only be modifed by
% changing the file.
%
% Input Parameters:
%
%   linear_constant: Double.  Negative values will correspond to systems
%   that are forward stable.  Non-negative values are required for the
%   viscosity solution of the HJ PDE to be well defined in backward
%   reachability.  Small values make the system more like a traditional
%   double integrator.  Optional.  Default is 1.0.
%
%   input_bounds: Two element non-negative double array.  Bounds on the
%   input signals in the (non-intuitive) order [ v, u ]. Optional.  Default
%   is [ 0.5, 0.25 ].
%
%   dynamics_term: String.  Specifies how the reach set should be
%   calculated.  Optional.  Default = 'lxf'.  Two options are available:
%
%     'combination': Use feedback inputs v = -sign(x_1) and u =
%     -sign(x_2), and then determine the reachable set using a combination
%     of forcing and discounting terms (for effects of x_1) and an upwinded
%     advection scheme (for x_2 dynamics).  Note that this choice of
%     feedback input is optimal for u but suboptimal for v.  Also, the
%     feedback control for v is non-monotonic in the solution value, so
%     it is not supported by viscosity solution theory.
%
%     'lxf': Use optimal feedback inputs as computed by general HJ PDE, and
%     then determine the reachable set using a Lax-Friedrichs scheme.  Note
%     that this approach will introduce more numerical dissipation.
%
% Output Parameters:
%
%   position_max: Vector, same size as the grid.  Maximum value in the
%   position coordinate which lies within the backward reach set at the end
%   of the integration.
%
%   position_min: Vector, same size as the grid.  Minimum value in the
%   position coordinate which lies within the backward reach set at the end
%   of the integration.
%
%   g: Grid structure on which position_max was computed.

% Copyright 2009 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Created by Ian Mitchell, 10/02/09
% Subversion tags for version control purposes.
% $Date: 2011-04-11 10:47:05 -0700 (Mon, 11 Apr 2011) $
% $Id: constantLinearMIE.m 581 2011-04-11 17:47:05Z mitchell $

  %---------------------------------------------------------------------------
  % You will see many executable lines that are commented out.
  %   These are included to show some of the options available; modify
  %   the commenting to modify the behavior.
  
  %---------------------------------------------------------------------------
  %% Make sure we can see the kernel m-files.
  run('addPathToKernel');

  %---------------------------------------------------------------------------
  %% Optional input parameters.

  if(nargin < 1)
    linear_constant = +1;
  end

  if(nargin < 2)
    input_bounds = [ +0.5, +0.25 ];
  end

  if(nargin < 3)
    dynamics_term = 'lxf';
  end

  %---------------------------------------------------------------------------
  %% PDE solver parameters.

  % Integration parameters.
  tMax = 4.0;                  % End time.
  plotSteps = 9;               % How many intermediate plots to produce?
  t0 = 0;                      % Start time.
  singleStep = 0;              % Plot at each timestep (overrides tPlot).

  % Period at which intermediate plots should be produced.
  tPlot = (tMax - t0) / (plotSteps - 1);

  % How close (relative) do we need to get to tMax to be considered finished?
  small = 100 * eps;

  % Pause after each plot?
  pauseAfterPlot = 0;

  % Delete previous plot before showing next?
  deleteLastPlot = 0;

  % Plot in separate subplots (set deleteLastPlot = 0 in this case)?
  useSubplots = 0;

  %---------------------------------------------------------------------------
  %% Optional and problem specific parameters.

  % Compute with both positive and negative velocities, or only positive?
  negative_velocities = 1;

  % Should we only allow the reach set to grow?
  restrict_update = 1;

  % Include the forcing function term?
  use_what = 'all';

  % Spatial approximation scheme.
  derivFunc = @upwindFirstENO2;

  % Temporal approximation scheme.
  integratorFunc = @odeCFL2;

  % Some integration parameters apply to any integrator.
  integratorOptions = odeCFLset('factorCFL', 0.9, 'stats', 'on');
  if(singleStep)
    integratorOptions = odeCFLset(integratorOptions, 'singleStep', 'on');
  end

  %---------------------------------------------------------------------------
  %% Create the grid.
  % Note that only the nonlinear variables need to be in the grid; in this
  % case, only one dimension.
  g.dim = 1;
  if negative_velocities
    g.min = 1.5 * [ -1 ]; %#ok<NBRAK>
  else
    g.min = 1.5 * [ 0 ]; %#ok<NBRAK,UNRCH>
  end
  g.max = 1.5 * [ +1 ]; %#ok<NBRAK>
  g.N = 151;
  g.bdry = @addGhostExtrapolate;
  g = processGrid(g);

  %---------------------------------------------------------------------------
  %% Create initial conditions.
  % A hexagon in velocity versus position space, which translates into
  % maximum (or minimum) position being a piecewise linear function of velocity.
  position_max = min(1.0, 1.5 - g.xs{1});
  position_max = min(position_max, g.xs{1} + 1.5);

  position_min = max(-1.0, g.xs{1} - 1.5);
  position_min = max(position_min, -g.xs{1} - 1.5);
  
  %---------------------------------------------------------------------------
  %% Set up PDE term that implements the nonlinear component of the dynamics.
  switch(dynamics_term)
    case 'combination'
      % I'm not sure if I am really guessing the feedback control for u1
      % totally wrong, or if I am implementing it wrong.  Either way, the
      % reachable set computed with this input is way different than the
      % set computed with LaxFriedrichs.
      warning('This option is buggy'); %#ok<WNTAG>
      % Given feedback inputs, a corresponding vector field can be
      % determined and implemented as a combination of discounting, forcing
      % and convectional flow field.

      %% Set up the forcing term in the PDE.
      % This term handles all of the dynamics of the linear state variables
      % EXCEPT the component that is linear in those variables.  We handle
      % the feedback input for v here, although it is not supported by
      % viscosity solution theory because it does not satisfy the degenerate
      % elliptic PDE monotonicity requirement. Remember to reverse the
      % dynamics for solving backward in time.
      forcingFunc = @termForcing;
      forcingData.grid = g;
      if(input_bounds(1) == 0)
        % No input v, so we can do a constant forcing term.
        forcingData.forcing = -(-g.xs{1});
      else
        % With an input the forcing depends on the solution which depends
        % on time, so we need to use a function.
        forcingData.forcing = @forcingFeedbackFunc;
        forcingData.grid = g;
        forcingData.input_bounds = input_bounds;
      end

      %% Set up the discounting term in the PDE.
      % This term handles the (linear) dynamics of the linear state variables.
      % Remember to reverse the dynamics for solving backward in time.
      discountFunc = @termDiscount;
      discountData.grid = g;
      discountData.lambda = -(-linear_constant);

      %% Set up the convection term in the PDE
      % This term handles the dynamics of the "nonlinear" variables.
      % Remember to reverse the dynamics for solving backward in time.
      nonlinearFunc = @termConvection;
      nonlinearData.grid = g;
      nonlinearData.derivFunc = derivFunc;
      nonlinearData.velocity = { -(-input_bounds(2) * sign(g.xs{1})) };

      %% Combine PDE terms if necessary.
      % These options are included for debugging purposes.  For proper
      % operation of the code, choose use_what = 'all'.
      switch(use_what)
        case 'none'
          schemeFunc = nonlinearFunc;
          schemeData = nonlinearData;
        case 'forcing'
          schemeFunc = @termSum;
          schemeData.innerFunc = { forcingFunc; nonlinearFunc };
          schemeData.innerData = { forcingData; nonlinearData };
        case 'discounting'
          schemeFunc = @termSum;
          schemeData.innerFunc = { discountFunc; nonlinearFunc };
          schemeData.innerData = { discountData; nonlinearData };
        case 'all'
          schemeFunc = @termSum;
          schemeData.innerFunc = { forcingFunc; discountFunc; nonlinearFunc };
          schemeData.innerData = { forcingData; discountData; nonlinearData };
        otherwise
          error('Unknown use_what value: %s', use_what);
      end

    %---------------------------------------------------------------------------
    case 'lxf'
      % A Lax-Friedrichs implementation can compute the optimal feedback
      % inputs although it is likely to be more dissipative numerically and
      % hence less accurate.
      schemeFunc = @termLaxFriedrichs;
      schemeData.derivFunc = derivFunc;
      schemeData.hamFunc = @hamFunc;
      schemeData.partialFunc = @partialFunc;
      schemeData.dissFunc = @artificialDissipationGLF;
      schemeData.grid = g;
      schemeData.linear_constant = linear_constant;
      schemeData.input_bounds = input_bounds;
      schemeData.use_what = use_what;
      % Set up a field which needs to be filled in during the actual
      % computation.
      schemeData.which_bound = 'undefined';
      
    otherwise
      error('Unknown dynamics_term: %s', dynamics_term);
  end

  %---------------------------------------------------------------------------
  %% Restrict the Hamiltonian so that reachable set only grows.
  % Whatever scheme handles the dynamics must already be completely set up.
  if restrict_update
    innerFunc = schemeFunc;
    innerData = schemeData;
    clear schemeFunc schemeData;

    % Wrap the true Hamiltonian inside the term approximation restriction routine.
    schemeFunc = @termRestrictUpdate;
    schemeData.innerFunc = innerFunc;
    schemeData.innerData = innerData;
    % Remember to set the parameter schemeData.positive below to the
    % appropriate value.
  end

  %---------------------------------------------------------------------------
  %% Initialize Display
  f = figure; %#ok<NASGU>

  % Set up subplot parameters if necessary.
  if useSubplots
    rows = ceil(sqrt(plotSteps));
    cols = ceil(plotSteps / rows);
    subplot(rows, cols, 1);
  end

  % To start with, we'll show both maximum and minimum position values, even
  % though we are going to compute them sequentially.
  h = show_function(g, position_min, [ 't = ' num2str(t0) ]);
  hold on;
  set(h, 'LineStyle', '--', 'LineWidth', 1, 'Color', 'b');
  h = show_function(g, position_max, [ 't = ' num2str(t0) ]);
  set(h, 'LineStyle', '--', 'LineWidth', 1, 'Color', 'b');
  xlabel('position');  ylabel('velocity');

  %---------------------------------------------------------------------------
  %% Integrate the HJ PDEs.
  % We have two (identical) PDEs to solve: one for the maximum position and
  % one for the minimum.  We could solve them together using vector level
  % sets, but for simplicity we'll just solve them sequentially.  
  %
  % I am generally loathe to use sub-functions, I'll use one here because the
  % relevant local scope is too big.  Remember that sub-functions use the
  % scope of their parent function, no matter where they are defined.
  % Therefore, you should never reuse a variable name unless you mean to
  % access the parent function's variable.

    %------------------------------------------------------------------------
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %------------------------------------------------------------------------
    function data = integrate_pde(data)
    % integrate_pde: A sub-function containing the integration loop.
    %
    %   data = integrate_pde(data)
    %
    % Just a copy of the standard ToolboxLS time integration loop.  All the
    % variables are taken from the parent function except data.  In
    % particular, you need to reset tNow and plotNum before each call if
    % you want to use this routine multiple times.
    % 
    % Parameters:
    %
    %   data: Array the same size as the grid, which is integrated from
    %   tNow until tMax (which are defined in the parent function).
    
      while(tMax - tNow > small * tMax)

        % Reshape data array into column vector for ode solver call.
        y0 = data(:);

        % How far to step?
        tSpan = [ tNow, min(tMax, tNow + tPlot) ];

        % Take a timestep.
        [ t y ] = feval(integratorFunc, schemeFunc, tSpan, y0,...
                        integratorOptions, schemeData);
        tNow = t(end);

        % Get back the correctly shaped data array
        data = reshape(y, g.shape);

        % Wait for last plot to be digested.
        if(pauseAfterPlot)
          pause;
        end

        % Delete last visualization if necessary.
        if(deleteLastPlot)
          delete(h);
        end

        % Move to next subplot if necessary.
        if(useSubplots)
          plotNum = plotNum + 1;
          subplot(rows, cols, plotNum);
        end

        % Create new visualization.
        h = show_function(g, data, [ 't = ' num2str(t) ]);
        set(h, 'LineStyle', ':', 'LineWidth', 1, 'Color', 'k');
        daspect([ 1 1 1 ]);

      end
    end
    %------------------------------------------------------------------------
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %------------------------------------------------------------------------

  startTime = cputime;

  % Solve the PDE for the maximum position.
  tNow = t0;
  plotNum = 1;
  if restrict_update
    % The maximum value should only decrease.
    schemeData.positive = 0;
    % Tell the hamFunc that we are working on the maximum value.
    schemeData.innerData.which_bound = 'maximum';
  else
    % Tell the hamFunc that we are working on the maximum value.
    schemeData.which_bound = 'maximum';     %#ok<UNRCH>
  end
  fprintf('\nIntegrating the maximum position values.');
  position_max = integrate_pde(position_max);

  % Solve the PDE for the minimum position.
  tNow = t0;
  plotNum = 1;
  if restrict_update
    % The minimum value should only increase.
    schemeData.positive = 1;
    % Tell the hamFunc that we are working on the maximum value.
    schemeData.innerData.which_bound = 'minimum';
  else
    % Tell the hamFunc that we are working on the maximum value.
    schemeData.which_bound = 'minimum';     %#ok<UNRCH>
  end
  fprintf('\nIntegrating the minimum position values.');
  position_min = integrate_pde(position_min);

  %---------------------------------------------------------------------------
  %% Termination code.  
  % Let the user know how long it all took.
  endTime = cputime;
  fprintf('Total execution time %g seconds\n', endTime - startTime);

  % Highlight the final reach set, including only those portions of the
  % maximum and minimum curves which are feasible (eg: minimum < maximum).
  not_feasible = (position_max < position_min);
  feasible_max = position_max;
  feasible_max(not_feasible) = NaN;
  feasible_min = position_min;
  feasible_min(not_feasible) = NaN;
  plot(feasible_max, g.xs{1}, 'LineWidth', 2, 'Color', 'red');
  plot(feasible_min, g.xs{1}, 'LineWidth', 2, 'Color', 'red');

  axis([ -1.5, +1.5, g.min, g.max ]);
  grid on;

    title(sprintf('MIE: t = %5.2f, a = %4.2f, v = %4.2f, u = %4.2f', ...
                  tNow, linear_constant, input_bounds(1), input_bounds(2)));

% Finished constantLinearMIE.
end

%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function h = show_function(g, data, title_string)
% show_function: A short helper function to display the approximate solution.
%
%   h = show_function(g, data, title)
%
% For this partially nonlinear reachability technique, the PDE solution is
% not an implicit surface function; therefore, the standard
% visualizeLevelSet() routine does not make sense.  This routine is supposed
% to emulate the simple calling functionality of visualizeLevelSet, but adapted
% appropriately for visualizing this particular problem.

  % For the double integrator, the grid coordinate is the velocity
  % (normally plotted as the vertical component) and the solution is the
  % maximum (or minimum) position (normally plotted as the horizontal
  % component.  Consequently, the parameters are reversed from their
  % typical order in the call to plot().
  h = plot(data, g.xs{1});
  title(title_string);
  
% Finished show_function.
end


%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function forcing = forcingFeedbackFunc(~, data, schemeData)
% forcingFeedbackFunc: Forcing term that includes feedback input.
%
%   forcing = forcingFeedbackFunc(~, data, schemeData)
%
% Implements the forcing function which implements the parts of the
% "linear" variable's dynamics that are not linear in itself.  In this
% case, it is the terms -x_2 - v, where the feedback input v =
% -sign(x_1) is used.  Note that this feedback does not satisfy
% monotonicity in the MIE formulation, so it is not supported by viscosity
% solution theory.
%
% Input parameters:
%
%   t: Time at beginning of timestep.
%
%   data: Data array.
%
%   schemeData: A structure (see below).
%
% Output parameters:
%
%   forcing: The forcing array.
%
% schemeData is a structure containing data specific to this Hamiltonian
% For this function it contains the field(s):
%
%   .grid: Grid structure.
%
%   .input_bounds: Two element non-negative double array.  Bounds on the
%   input signals.

% Ian Mitchell 2010/09/07

  % Two terms: drift due to x_2 (velocity) and feedback input v =
  % -sign(x_1), where in the MIE framework x_1 is the solution.  Remember to
  % reverse the dynamics (multiply by -1) since we solve the PDE backward
  % in time.
  forcing = -(-schemeData.grid.xs{1} - (-schemeData.input_bounds(1) * sign(data)));

end

%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function hamValue = hamFunc(~, data, deriv, schemeData)
% hamFunc: analytic Hamiltonian.
%
% hamValue = hamFunc(t, data, deriv, schemeData)
%
% This function implements the hamFunc prototype for the MIE reachability
% version of the constant linear with input variation on the double
% integrator.
%
% Input parameters:
%
%   t: Time at beginning of timestep.
%
%   data: Data array.
%
%   deriv: Cell vector of the costate (\grad \phi).
%
%   schemeData: A structure (see below).
%
% Output parameters:
%
%   hamValue: The analytic hamiltonian.
%
% schemeData is a structure containing data specific to this Hamiltonian
% For this function it contains the field(s):
%
%   .grid: Grid structure.
%
%   .linear_constant: Double.  Constant multiplier for the linear term of
%   the integrator.
%
%   .input_bounds: Two element non-negative double array.  Bounds on the
%   input signals.
%
%   .use_what: String.
%
%   .which_bound: String.  Specifies whether the computation is working on
%   the maximum or the minimum value of x_2 (position).  There are two
%   valid options: 'maximum' or 'minimum'.

% Ian Mitchell 2010/09/07

  checkStructureFields(schemeData, 'grid', 'linear_constant', 'input_bounds', ...
                                   'use_what', 'which_bound');

  % We handle the terms in the following order: dynamics of "nonlinear"
  % variable (x_2), linear dynamics of "linear" variable (x_1), other
  % dynamics of "linear" variable (which includes dynamics due to
  % "nonlinear" variables and any inputs). Remember to reverse the dynamics
  % (multiply by -1) since we solve the PDE backward in time.

  switch(schemeData.which_bound)
    case 'maximum'
      u1_input = -schemeData.input_bounds(1);
      u2_input = +abs(deriv{1}) * schemeData.input_bounds(2);
    case 'minimum'
      u1_input = +schemeData.input_bounds(1);
      u2_input = -abs(deriv{1}) * schemeData.input_bounds(2);      
    otherwise
      error('Unknown which_bound field %s', schemeData.which_bound);
  end
  
  switch(schemeData.use_what)
    case 'none'
      hamValue = -(u2_input);
    case 'forcing'
      hamValue = -(u2_input - schemeData.grid.xs{1} - u1_input);
    case 'discounting'
      hamValue = -(u2_input - schemeData.linear_constant * data);
    case 'all'
      hamValue = -(u2_input - schemeData.linear_constant * data ...
                   - schemeData.grid.xs{1} - u1_input);
    otherwise
      error('Unknown use_what value: %s', use_what);
  end

end % function hamFunc().


%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function alpha = partialFunc(~, ~, ~, ~, schemeData, dim)
% partialFunc: Hamiltonian partial function.
%
% alpha = partialFunc(t, data, derivMin, derivMax, schemeData, dim)
%
% This function implements the partialFunc prototype for the MIE
% reachability version of the constant linear with input variation on the
% double integrator.
%
% It calculates the extrema of the absolute value of the partials of the 
% analytic Hamiltonian with respect to the costate (gradient).
%
% Parameters:
%
%   t: Time at beginning of timestep.
%
%   data: Data array.
%
%   derivMin: Cell vector of minimum values of the costate (\grad \phi).
%
%   derivMax: Cell vector of maximum values of the costate (\grad \phi).
%
%   schemeData: A structure (see below).
%
%   dim: Dimension in which the partial derivatives is taken.
%
% Output Parameters:
%
%   alpha: Maximum absolute value of the partial of the Hamiltonian with
%   respect to the costate in dimension dim for the specified range of
%   costate values (O&F equation 5.12). Note that alpha can (and should) be
%   evaluated separately at each node of the grid.
%
% schemeData is a structure containing data specific to this Hamiltonian
% For this function it contains the field(s):
%
%   .grid: Grid structure.
%
%   .input_bounds: Two element non-negative double array.  Bounds on the
%   input signals.

% Ian Mitchell 2010/09/07

  checkStructureFields(schemeData, 'grid', 'input_bounds');

  % Very few terms in the Hamiltonian actually depend on the costate, so
  % this partials function is very simple to compute.
  switch dim
    case 1
      alpha = schemeData.input_bounds(2);

    otherwise
      error('Partials do not exist in dimension %d', dim);
  end
  
end % function partialFunc().
