function [ position_max, position_min, g ] = doubleIntegratorMIE()
% doubleIntegratorMIE: double integrator reachability (1D MIE scheme)
%
%   [ position_max, position_min, g ] = doubleIntegratorMIE()
%  
% This is a reachability calculation for the double integrator using the
% Mixed Implicit Explicit scheme.  Note the nonstandard variable
% definition for the double integrator variables because we want to agree
% with the variable definitions for the partially nonlinear reachability
% paper:
%
%     \dot x = u   (velocity; the vertical dimension)
%     \dot y = x   (position; the horizontal dimension)
%        with input |u| \leq U and U is some constant.
%
% The y variable is the "linear" variable which is treated specially -- its
% maximum (or minimum) value is tracked as the solution to a HJ PDE in the
% "nonlinear" x variables.
%
% Note that position_max and position_min are defined everywhere in the
% computational domain; however, only position values that lie in the
% interval [ position_min, position_max ] are within the reach set.
% Notably, if position_min > position_max for some points in the
% computational domain, then the reach set is empty at those points.
%
% There are many options in the function which can only be modifed by
% changing the file.
%
% Input Parameters: none.
%
% Output Parameters:
%
%   position_max: Vector, same size as the grid.  Maximum value in the
%   position coordinate which lies within the backward reach set at the end
%   of the integration.
%
%   position_min: Vector, same size as the grid.  Minimum value in the
%   position coordinate which lies within the backward reach set at the end
%   of the integration.
%
%   g: Grid structure on which position_max was computed.

% Copyright 2009 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Created by Ian Mitchell, 9/01/09
% Subversion tags for version control purposes.
% $Date: 2011-04-11 10:47:05 -0700 (Mon, 11 Apr 2011) $
% $Id: doubleIntegratorMIE.m 581 2011-04-11 17:47:05Z mitchell $

  %---------------------------------------------------------------------------
  % You will see many executable lines that are commented out.
  %   These are included to show some of the options available; modify
  %   the commenting to modify the behavior.
  
  %---------------------------------------------------------------------------
  %% Make sure we can see the kernel m-files.
  run('addPathToKernel');

  %---------------------------------------------------------------------------
  %% PDE solver parameters.

  % Integration parameters.
  tMax = 4.0;                  % End time.
  plotSteps = 9;               % How many intermediate plots to produce?
  t0 = 0;                      % Start time.
  singleStep = 0;              % Plot at each timestep (overrides tPlot).

  % Period at which intermediate plots should be produced.
  tPlot = (tMax - t0) / (plotSteps - 1);

  % How close (relative) do we need to get to tMax to be considered finished?
  small = 100 * eps;

  % Pause after each plot?
  pauseAfterPlot = 0;

  % Delete previous plot before showing next?
  deleteLastPlot = 0;

  % Plot in separate subplots (set deleteLastPlot = 0 in this case)?
  useSubplots = 0;

  %---------------------------------------------------------------------------
  %% Optional and problem specific parameters.

  % Maximum magnitude of the input.
  input_bound = 0.25;

  % Compute with both positive and negative velocities, or only positive?
  negative_velocities = 1;

  % Should we only allow the reach set to grow?
  restrict_update = 1;

  % Include the forcing function term?
  use_forcing = 1;

  % How to implement the dynamics in the HJ PDE?
  dynamics_term = 'convection';
  % Could also implement as Lax-Freidrichs...

  % Spatial approximation scheme.
  derivFunc = @upwindFirstENO2;

  % Temporal approximation scheme.
  integratorFunc = @odeCFL2;

  % Some integration parameters apply to any integrator.
  integratorOptions = odeCFLset('factorCFL', 0.9, 'stats', 'on');
  if(singleStep)
    integratorOptions = odeCFLset(integratorOptions, 'singleStep', 'on');
  end

  %---------------------------------------------------------------------------
  %% Create the grid.
  % Note that only the nonlinear variables need to be in the grid; in this
  % case, only one dimension.
  g.dim = 1;
  if negative_velocities
    g.min = 1.5 * [ -1 ]; %#ok<*NBRAK>
  else
    g.min = 1.5 * [ 0 ]; %#ok<UNRCH>
  end
  g.max = 1.5 * [ +1 ];
  g.N = 101;
  g.bdry = @addGhostExtrapolate;
  g = processGrid(g);

  %---------------------------------------------------------------------------
  %% Create initial conditions.
  % A hexagon in velocity versus position space, which translates into
  % maximum (or minimum) position being a piecewise linear function of velocity.
  position_max = min(1.0, 1.5 - g.xs{1});
  position_max = min(position_max, g.xs{1} + 1.5);

  position_min = max(-1.0, g.xs{1} - 1.5);
  position_min = max(position_min, -g.xs{1} - 1.5);
  
  %---------------------------------------------------------------------------
  %% Set up PDE term that implements the nonlinear component of the dynamics.
  switch(dynamics_term)
    case 'convection'
      % The optimal feedback input for the double integrator can be determined 
      % analytically and thus the optimal dynamics can be implemented as a
      % purely convectional flow field.  Remember to reverse the dynamics
      % since we solve the PDE backward in time.
      nonlinearFunc = @termConvection;
      nonlinearData.grid = g;
      nonlinearData.derivFunc = derivFunc;
      nonlinearData.velocity = { -(-input_bound * sign(g.xs{1})) };

    %case 'LaxFriedrichs'
      % A Lax-Friedrichs implementation would be more representative of what
      % would be used for more complicated dynamics, although it is likely to
      % be more dissipative and hence less accurate.

    otherwise
      error('Unknown dynamics_term: %s', dynamics_term);
  end

  %---------------------------------------------------------------------------
  %% Set up the forcing term in the PDE.
  % This term handles the dynamics of the linear state variables.  In this
  % case, the forcing function is (the negative of) the same function as the
  % dynamics of the second "nonlinear" coordinate (but note that the effect
  % is very different, since the forcing function is directly added to the
  % PDE, while the nonlinear dynamics are multiplied by the function
  % derivative).  Remember to reverse the dynamics for solving backward in
  % time.
  forcingFunc = @termForcing;
  forcingData.grid = g;
  forcingData.forcing = -(-g.xs{1});

  %---------------------------------------------------------------------------
  %% Combine PDE terms if necessary.
  if use_forcing
    schemeFunc = @termSum;
    schemeData.innerFunc = { forcingFunc; nonlinearFunc };
    schemeData.innerData = { forcingData; nonlinearData };
  else
    schemeFunc = nonlinearFunc; %#ok<UNRCH>
    schemeData = nonlinearData;
  end

  %---------------------------------------------------------------------------
  %% Restrict the Hamiltonian so that reachable set only grows.
  % Whatever scheme handles the dynamics must already be completely set up.
  if restrict_update
    innerFunc = schemeFunc;
    innerData = schemeData;
    clear schemeFunc schemeData;

    % Wrap the true Hamiltonian inside the term approximation restriction routine.
    schemeFunc = @termRestrictUpdate;
    schemeData.innerFunc = innerFunc;
    schemeData.innerData = innerData;
    % Remember to set the parameter schemeData.positive below to the
    % appropriate value.
  end

  %---------------------------------------------------------------------------
  %% Initialize Display
  f = figure; %#ok<NASGU>

  % Set up subplot parameters if necessary.
  if useSubplots
    rows = ceil(sqrt(plotSteps));
    cols = ceil(plotSteps / rows);
    subplot(rows, cols, 1);
  end

  % To start with, we'll show both maximum and minimum position values, even
  % though we are going to compute them sequentially.
  h = show_function(g, position_min, [ 't = ' num2str(t0) ]);
  hold on;
  set(h, 'LineStyle', '--', 'LineWidth', 1, 'Color', 'b');
  h = show_function(g, position_max, [ 't = ' num2str(t0) ]);
  set(h, 'LineStyle', '--', 'LineWidth', 1, 'Color', 'b');
  xlabel('position');  ylabel('velocity');

  %---------------------------------------------------------------------------
  %% Integrate the HJ PDEs.
  % We have two (identical) PDEs to solve: one for the maximum position and
  % one for the minimum.  We could solve them together using vector level
  % sets, but for simplicity we'll just solve them sequentially.  
  %
  % I am generally loathe to use sub-functions, I'll use one here because the
  % relevant local scope is too big.  Remember that sub-functions use the
  % scope of their parent function, no matter where they are defined.
  % Therefore, you should never reuse a variable name unless you mean to
  % access the parent function's variable.

    %------------------------------------------------------------------------
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %------------------------------------------------------------------------
    function data = integrate_pde(data)
    % integrate_pde: A sub-function containing the integration loop.
    %
    %   data = integrate_pde(data)
    %
    % Just a copy of the standard ToolboxLS time integration loop.  All the
    % variables are taken from the parent function except data.  In
    % particular, you need to reset tNow and plotNum before each call if
    % you want to use this routine multiple times.
    % 
    % Parameters:
    %
    %   data: Array the same size as the grid, which is integrated from
    %   tNow until tMax (which are defined in the parent function).
    
      while(tMax - tNow > small * tMax)

        % Reshape data array into column vector for ode solver call.
        y0 = data(:);

        % How far to step?
        tSpan = [ tNow, min(tMax, tNow + tPlot) ];

        % Take a timestep.
        [ t y ] = feval(integratorFunc, schemeFunc, tSpan, y0,...
                        integratorOptions, schemeData);
        tNow = t(end);

        % Get back the correctly shaped data array
        data = reshape(y, g.shape);

        % Wait for last plot to be digested.
        if(pauseAfterPlot)
          pause;
        end

        % Delete last visualization if necessary.
        if(deleteLastPlot)
          delete(h);
        end

        % Move to next subplot if necessary.
        if(useSubplots)
          plotNum = plotNum + 1;
          subplot(rows, cols, plotNum);
        end

        % Create new visualization.
        h = show_function(g, data, [ 't = ' num2str(t) ]);
        set(h, 'LineStyle', ':', 'LineWidth', 1, 'Color', 'k');
        daspect([ 1 1 1 ]);

      end
    end
    %------------------------------------------------------------------------
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %------------------------------------------------------------------------

  startTime = cputime;

  % Solve the PDE for the maximum position.
  tNow = t0;
  plotNum = 1;
  if restrict_update
    % The maximum value should only decrease.
    schemeData.positive = 0;
  end
  fprintf('\nIntegrating the maximum position values.');
  position_max = integrate_pde(position_max);

  % Solve the PDE for the minimum position.
  tNow = t0;
  plotNum = 1;
  if restrict_update
    % The minimum value should only increase.
    schemeData.positive = 1;
  end
  fprintf('\nIntegrating the minimum position values.');
  position_min = integrate_pde(position_min);

  %---------------------------------------------------------------------------
  %% Termination code.  
  % Let the user know how long it all took.
  endTime = cputime;
  fprintf('Total execution time %g seconds\n', endTime - startTime);

  % Highlight the final reach set, including only those portions of the
  % maximum and minimum curves which are feasible (eg: minimum < maximum).
  not_feasible = (position_max < position_min);
  % Reduce the not feasible set by one node on either side so that the
  % resulting curves overlap.
  not_feasible = not_feasible & [ true; not_feasible(1:end-1) ] & [ not_feasible(2:end); true ];
  feasible_max = position_max;
  feasible_max(not_feasible) = NaN;
  feasible_min = position_min;
  feasible_min(not_feasible) = NaN;
  plot(feasible_max, g.xs{1}, 'LineWidth', 2, 'Color', 'red');
  plot(feasible_min, g.xs{1}, 'LineWidth', 2, 'Color', 'red');

  axis([ -1.5, +1.5, g.min, g.max ]);
  title(sprintf('Double Integrator with MIE reachability at t = %4.2f with |u| \\leq %4.2f', tMax, input_bound));
  grid on;
  
% Finished doubleIntegratorMIE.
end

%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function h = show_function(g, data, title_string)
% show_function: A short helper function to display the approximate solution.
%
%   h = show_function(g, data, title)
%
% For this partially nonlinear reachability technique, the PDE solution is
% not an implicit surface function; therefore, the standard
% visualizeLevelSet() routine does not make sense.  This routine is supposed
% to emulate the simple calling functionality of visualizeLevelSet, but adapted
% appropriately for visualizing this particular problem.

  % For the double integrator, the grid coordinate is the velocity
  % (normally plotted as the vertical component) and the solution is the
  % maximum (or minimum) position (normally plotted as the horizontal
  % component.  Consequently, the parameters are reversed from their
  % typical order in the call to plot().
  h = plot(data, g.xs{1});
  title(title_string);
  
% Finished show_function.
end
