function [ solution, g ] = obliviousMIE()
% obliviousMIE: Two vehicles model, but only pursuer has inputs.
%
%   [ solution, g ] = obliviousMIE()
%
% Mixed Implicit-Explicit reachability calculation for a restricted version
% of the two identical vehicles model: Assume that the evader has fixed
% heading and velocity, but the pursuer can adjust its linear acceleration
% and angular velocity.  Position variables x and y are the terminal
% integrators.
%
% The model is:
%
%     \dot x = -v_e + v_p \cos \psi
%     \dot y = v_p \sin \psi
%     \dot v_p = a_p
%     \dot \psi = \omega_p
%
% where x, y, psi are relative coordinates with the evader fixed at the
% origin facing the positive x-axis, while v_p and v_e are the forward
% speed of the pursuer and evader respectively.  Inputs are pursuer linear
% acceleration and angular velocity.
%
% The target set is a unit box in x - y space, subject to the pursuer's
% linear velocity being within some interval.  We consider pursuer linear
% velocities that are too high or too low to be part of an escape set.
% Note that the evader has no inputs, and so cannot drive the system into
% this escape set consciously.  Instead, the pursuer must use its inputs to
% avoid the escape set.
%
% It would be easy to add evader linear acceleration (and make evader
% linear velocity a state variable) except that it would make visualization
% hard and it is not clear how to handle minimum and maximum speeds for the
% evader.  (Maybe part of the target set?)
%
% There are many options in the function which can only be modifed by
% changing the file.
%
% Input Parameters: None.
%
% Output Parameters:
%
%   solution: A structure with four fields.  Each field is an array
%   representing a grid function; in this case, a function over the gridded
%   variables psi and v_p.  The functions are the upper and lower bounds of
%   the reachable set for the position variables x and y.  The fields in
%   the structure are .x_max, .x_min, .y_max, .y_min with the obvious
%   interpretations.
%
%   g: Grid structure on which the solution functions were computed.

% Copyright 2011 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Created by Ian Mitchell, 2011/01/26.
% Subversion tags for version control purposes.
% $Date: 2011-04-12 20:10:31 -0500 (Tue, 12 Apr 2011) $
% $Id: obliviousMIE.m 582 2011-04-13 01:10:31Z mitchell $

  %---------------------------------------------------------------------------
  % You will see many executable lines that are commented out.
  %   These are included to show some of the options available; modify
  %   the commenting to modify the behavior.
  
  %---------------------------------------------------------------------------
  %% Make sure we can see the kernel m-files.
  run('addPathToKernel');

  %---------------------------------------------------------------------------
  %% Optional input parameters.

  %---------------------------------------------------------------------------
  %% PDE solver parameters.

  % Integration parameters.
  tMax = 2.0;                  % End time.
  plotSteps = 9;               % How many intermediate plots to produce?
  t0 = 0;                      % Start time.
  singleStep = 0;              % Plot at each timestep (overrides tPlot).

  % Period at which intermediate plots should be produced.
  tPlot = (tMax - t0) / (plotSteps - 1);

  % How close (relative) do we need to get to tMax to be considered finished?
  small = 100 * eps;

  % Pause after each plot?
  pauseAfterPlot = false;

  % Should we only allow the reach set to grow?
  restrict_update = true;
  
  % Should we mask after every timestep to maintain the restrictions on
  % pursuer velocity?
  use_mask = true;

  % Spatial approximation scheme.
  derivFunc = @upwindFirstENO2;

  % Temporal approximation scheme.
  integratorFunc = @odeCFL2;

  % Some integration parameters apply to any integrator.
  integratorOptions = odeCFLset('factorCFL', 0.9, 'stats', 'on');
  if(singleStep)
    integratorOptions = odeCFLset(integratorOptions, 'singleStep', 'on');
  end
  
  % When implementing the initial conditions, what slope should we give to
  % the portions of the function which should be discontinuities if we
  % could implement the true box target set? Larger values are closer to a
  % true box, but the result will behave more like a discontinuity in the
  % MIE version.  Note that this particular number may not be the actual
  % slope used.
  modified_box_slope_factor = 5;

  % Properties of the surface plots for visualizing the reachable set.
  surf_plot_properties = { 'FaceColor', 'red', 'EdgeColor', 'None' };
  
  %---------------------------------------------------------------------------
  %% Optional and problem specific parameters.

  % These parameters were chosen primarily to give a reach set which shows
  % significant differentiation over the range of psi and vp, as well as
  % significant difference from the target set.  I also tried to choose
  % parameters which would lead to a relatively well-scaled grid (eg: g.dx
  % does not vary too much from dimension to dimension), and keep the reach
  % set away from the computational boundaries.
  
  % The target "box" in state space.
  params.target_min = [ -1; -1; 1.0; -inf ];
  params.target_max = [ +1; +1; 3.0; +inf ];
  % The mask which represents the set of states (essentially velocities)
  % outside of which the pursuer may not go to achieve the target set.  We
  % implement this bound as a mask because it must be enforced throughout
  % the simulation, rather than just at the initial time.
  params.mask_min = [ -inf; -inf; 1.0; -inf ];
  params.mask_max = [ +inf; +inf; 3.0; +inf ];
  % Bounds on pursuer linear acceleration.
  params.linear_acc_p = [ -0.2, +0.2 ];
  % Bounds on pursuer angular velocity.
  params.angular_vel_p = [ -0.2, +0.2 ];
  % Evader's linear velocity (fixed).
  params.linear_vel_e = +1;
  % How should the inputs be handled for the different components of the
  % vector solution?
  params.input_choice = 'independent';
  %params.input_choice = 'sum';
  
  % For clarity of references in the code, create variables mapping from
  % state names to indexes and vice versa.
  s2i = struct('vp', 1, 'psi', 2);
  i2s = { 'vp', 'psi' };
  % Also a structure mapping from state names to the indexes in grid_all,
  % target_min, target_max, mask_min, mask_max.
  all_s2i = struct('x', 1, 'y', 2, 'vp', 3, 'psi', 4);

  %---------------------------------------------------------------------------
  %% Create the grid.
  % Note that only the nonlinear variables need to be in the grid; in this
  % case, only one dimension.
  g.dim = 2;
  g.min = [ 0.8;    0 ];
  g.max = [ 3.2; 2*pi ];
  % For non-periodic dimensions, we want a round number + 1, while for
  % periodic dimensions we want just the round number.
  g.N = 1 * [ 32; 50 ] + [ 1; 0 ];
  % Adjust location of upper-most node in grid for periodic coordinate.
  g.max(s2i.psi) = g.max(s2i.psi) * (1 - 1 / g.N(s2i.psi));
  g.bdry = { @addGhostExtrapolate; @addGhostPeriodic };
  g = processGrid(g);

  %---------------------------------------------------------------------------
  %% Create initial and masking conditions.

  solution.x_max = min(params.target_max(all_s2i.x) * ones(g.shape), ...
                       min(-modified_box_slope_factor * (g.xs{s2i.vp} - params.target_max(all_s2i.vp)), ...
                           +modified_box_slope_factor * (g.xs{s2i.vp} - params.target_min(all_s2i.vp))));
  solution.x_min = max(params.target_min(all_s2i.x) * ones(g.shape), ...
                       max(+modified_box_slope_factor * (g.xs{s2i.vp} - params.target_max(all_s2i.vp)), ...
                           -modified_box_slope_factor * (g.xs{s2i.vp} - params.target_min(all_s2i.vp))));
  solution.y_max = min(params.target_max(all_s2i.y) * ones(g.shape), ...
                       min(-modified_box_slope_factor * (g.xs{s2i.vp} - params.target_max(all_s2i.vp)), ...
                           +modified_box_slope_factor * (g.xs{s2i.vp} - params.target_min(all_s2i.vp))));
  solution.y_min = max(params.target_min(all_s2i.y) * ones(g.shape), ...
                       max(+modified_box_slope_factor * (g.xs{s2i.vp} - params.target_max(all_s2i.vp)), ...
                           -modified_box_slope_factor * (g.xs{s2i.vp} - params.target_min(all_s2i.vp))));
  
  % How are we going to pack these four solutions into a single array?
  % Pack them in the same order in which they will appear in subplots to
  % make the visualization code shorter.
  c2i = struct('x_max', 1, 'y_max', 2, 'x_min', 3, 'y_min', 4);
  i2c = { 'x_max', 'y_max', 'x_min', 'y_min' };
  % How many elements do we have in our vector solution?
  vector_length = length(i2c);

  % Because we no longer have the luxury of an implicit surface
  % representation, we cannot build a continuous masking function.
  % The masking function will be infinity for velocities within the mask
  % range and zero for those outside.
  in_mask_range = ((g.xs{s2i.vp} <= params.mask_max(all_s2i.vp)) ...
                    & (g.xs{s2i.vp} >= params.mask_min(all_s2i.vp)));
  mask.x_max = zeros(g.shape);
  mask.x_max(in_mask_range) = +inf;
  mask.x_min = zeros(g.shape);
  mask.x_min(in_mask_range) = -inf;
  mask.y_max = mask.x_max;
  mask.y_min = mask.x_min;
  
  % Ensure that the initial conditions satisfy the mask.
  for i = 1 : vector_length
    switch i2c{i}
      case { 'x_max', 'y_max' }
        solution.(i2c{i}) = min(solution.(i2c{i}), mask.(i2c{i}));
      case { 'x_min', 'y_min' }
        solution.(i2c{i}) = max(solution.(i2c{i}), mask.(i2c{i}));
      otherwise
        error('Unknown solution name %s', i2c{i});
    end
  end

  %---------------------------------------------------------------------------
  %% Set up PDE term that implements the nonlinear component of the dynamics.

  % Basic setup for Lax-Friedrichs scheme.
  schemeFunc = @termVectorLaxFriedrichs;
  schemeData.derivFunc = derivFunc;
  schemeData.hamFunc = @hamFunc;
  schemeData.partialFunc = @partialFunc;
  schemeData.dissFunc = @artificialDissipationGLF;
  schemeData.grid = g;

  % We cannot use the regular termRestrictUpdate routine since some
  % elements of the vector must not grow (the minimums) and some must not
  % shrink (the maximums).  Consequently, we need to do the restriction
  % inside the Hamiltonian.
  schemeData.restrict_update = restrict_update;
  
  % Problem dependent parameters.
  schemeData.params = params;
  schemeData.s2i = s2i;
  schemeData.i2s = i2s;
  schemeData.c2i = c2i;
  schemeData.i2c = i2c;
  schemeData.vector_length = vector_length;

  % Some additional values which we can precompute and hence save time.
  
  % We need a cell vector of indexes that refers to an entire grid array so
  % that we can unpack the data array both inside and outside the
  % Hamiltonian.
  schemeData.grid_index = cell(g.dim, 1);
  for i = 1 : g.dim
    schemeData.grid_index{i} = 1 : g.N(i);
  end

  % Some stuff related to the effect of heading on dynamics.
  schemeData.dxdt = -params.linear_vel_e + g.xs{s2i.vp} .* cos(g.xs{s2i.psi});
  schemeData.dydt = g.xs{s2i.vp} .* sin(g.xs{s2i.psi});

  %---------------------------------------------------------------------------
  %% Set up the masking.
  % We cannot use the regular masking routine because some elements
  % are masked from above and some from below.

  if use_mask
    % Note that the mask data for MIE calculations is not in the same
    % format as regular masking functions.
    schemeData.mask_data = mask;
    integratorOptions = odeCFLset(integratorOptions, 'postTimestep', @postTimestepMaskMIE);
  end
  
  %---------------------------------------------------------------------------
  %% Initialize Display
  f = figure;

  % We plot the position variables (terminal integrators) vertically
  % because that is easy to do with surf().  We also use the naming
  % convention from the HSCC paper.
  for i = 1 : vector_length
    cmpt = i2c{i};
    subplot(2,2,i)
    surf(g.xs{:}, solution.(cmpt), surf_plot_properties{:});
    prettyPlot(cmpt);
  end
  
  %---------------------------------------------------------------------------
  %% Main loop

  % Loop until tMax (subject to a little roundoff).
  tNow = t0;
  startTime = cputime;
  while(tMax - tNow > small * tMax)

    % Pack the elements of the vector solution into a single array.  We
    % could use struct2cell, but then we don't have any control over the
    % ordering of the fields.
    solution_cell = cell(vector_length, 1);
    for i = 1 : vector_length
      solution_cell{i} = solution.(i2c{i});
    end
    data = cat(g.dim + 1, solution_cell{:});
    
    % Reshape data array into column vector for ode solver call.
    y0 = data(:);

    % How far to step?
    tSpan = [ tNow, min(tMax, tNow + tPlot) ];

    % Take a timestep.
    [ t y ] = feval(integratorFunc, schemeFunc, tSpan, y0,...
                    integratorOptions, schemeData);
    tNow = t(end);

    % Get back the correctly shaped data array.  We need to use grid.N here
    % instead of grid.shape because for a 1D grid we use the second
    % dimension to store the components.
    data = reshape(y, [ g.N', vector_length ]);
    
    % Unpack the elements of the vector solution.
    for i = 1 : vector_length
      solution.(i2c{i}) = data(schemeData.grid_index{:}, i);
    end
    
    if(pauseAfterPlot)
      % Wait for last plot to be digested.
      pause;
    end

    % Return to last plot if necessary.
    figure(f);

    % Visualize the current solution.
    for i = 1 : vector_length
      cmpt = i2c{i};
      subplot(2,2,i)
      cla;
      surf(g.xs{:}, solution.(cmpt), surf_plot_properties{:});
      prettyPlot(cmpt);
    end
    
  end
  
  %---------------------------------------------------------------------------
  %% Termination code.  
  % Let the user know how long it all took.
  endTime = cputime;
  fprintf('Total execution time %g seconds\n', endTime - startTime);
  % Put a title over all subplots.  If you want to uncomment the following
  % line, get mtit() from Matlab Central.
  %mtit(sprintf('Oblivious evader model with MIE reachability at t = %4.1f', tNow));

end  % Main function.


%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function prettyPlot(z_axis_label)
% prettyPlot: Make the plots of the reachable set look good.

  xlabel('vp');
  ylabel('psi');
  h_zlabel = zlabel(z_axis_label);
  % Turn off the tex interpreter so that underscore characters are not
  % interpreted to be subscripts.
  set(h_zlabel, 'Interpreter', 'None');
  camlight headlight;
  grid on;
  axis tight;
end % prettyPlot().

%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function [ y, schemeData ] = postTimestepMaskMIE(~, y, schemeData)
% postTimestepMaskMIE: Masking function specialized for this MIE calculation.
%
%  [ y, schemeData ] = postTimestepMask(t, y, schemeData)
%
% This routine implements the postTimestepFunc prototype for masking the
% solution of an MIE reachability calculation.  Because the MIE solution
% includes multiple minimum and maximum components, some elements are
% masked to only grow and some to only shrink.
%
% Input Parameters:
%
%   t: Current time (ignored).
%
%   y: Input version of the level set function, in vector form.
%
%   schemeData: Input version of a structure (see below).
%
% Output Parameters:
%
%   y: Output version of the level set function, in vector form.
%
%   schemeData: Output version of the structure (unmodified).
%
% The structure schemeData contains data specific to this type of term
% approximation.  For this function it contains the field(s):
%
%   .grid: Grid structure.
%
%   .i2c: Cell vector.  Mapping from indexes in the extra dimension of the
%   data array to names for the components of the vector HJ PDE.
%
%   .vector_length: Integer.  Number of components in the vector HJ PDE.
%
%   .grid_index: Cell vector.  Contains an index cell vector for every
%   element of a data array corresponding to the grid.
%
%   .mask_data: A structure containing the mask data for each component of
%   the solution.  The structure field names should be the same as the
%   component names in i2c.
%
% schemeData may contain other fields.

  % Unpack the problem data that we use a lot into variables with short
  % names.
  i2c = schemeData.i2c;
  vector_length = schemeData.vector_length;
  grid = schemeData.grid;
  
  %------------------------------------------------------------------------
  %% Unpack the data.
  data = reshape(y, [ grid.N', vector_length ]);

  % Unpack the elements of the vector solution.
  for i = 1 : vector_length
    solution.(i2c{i}) = data(schemeData.grid_index{:}, i);
  end
  
  %------------------------------------------------------------------------
  %% Apply the masking.  
  for i = 1 : vector_length
    cmpt = i2c{i};
    % Use the last three letters of the component name to decide what
    % operation to apply to the component.
    switch(cmpt(end-2:end));
      case 'max'
        % Solution components associated with maxima must be below their mask.
        solution.(cmpt) = min(solution.(cmpt), schemeData.mask_data.(cmpt));
      case 'min'
        % Solution components associated with minima must be above their mask.
        solution.(cmpt) = max(solution.(cmpt), schemeData.mask_data.(cmpt));
      otherwise
        error('Not known how to mask solution component: %s', cmpt);
    end
  end
  
  %------------------------------------------------------------------------
  %% Repack the data.
  % Copied from main loop.
  solution_cell = cell(vector_length, 1);
  for i = 1 : vector_length
    solution_cell{i} = solution.(i2c{i});
  end
  data = cat(grid.dim + 1, solution_cell{:});
  
  % Reshape data array into column vector for ode solver call.
  y = data(:);

end % postTimestepMaskMIE();

%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function ham_value = hamFunc(~, ~, deriv, schemeData)
% hamFunc: analytic Hamiltonian.
%
%   ham_value = hamFunc(t, data, deriv, schemeData)
%
% This function implements the hamFunc prototype for the MIE reachability
% version of the oblivious two vehicle problem (eg: evader has no inputs).
%
% Input parameters:
%
%   t: Time at beginning of timestep.
%
%   data: Data array.  Note that it contains one more dimension than the
%   underlying grid.  This extra dimension is used to pack the components
%   of the vector HJ PDE that we are solving.  See termVectorLaxFriedrichs() 
%   for details.
%
%   deriv: Cell vector of the costate (\grad \phi).
%
%   schemeData: A structure (see below).
%
% Output parameters:
%
%   hamValue: The analytic hamiltonian.
%
% schemeData is a structure containing data specific to this Hamiltonian
% For this function it contains the field(s):
%
%   .grid: Grid structure.
%
%   .params: Model parameters.
%
%   .s2i: Structure.  Maps from variable names to the corresponding
%   grid dimension (or false if that variable does not appear in the grid).
%
%   .i2c: Cell vector.  Mapping from indexes in the extra dimension of the
%   data array to names for the components of the vector HJ PDE.
%
%   .vector_length: Integer.  Number of components in the vector HJ PDE.
%
%   .grid_index: Cell vector.  Contains an index cell vector for every
%   element of a data array corresponding to the grid.
%
%   .dxdt, .dydt: Precomputed dynamics for position variables.

% Ian Mitchell 2011/01/26

  checkStructureFields(schemeData, 'grid', 'params', 's2i', ...
                                   'i2c', 'vector_length', 'grid_index', ...
                                   'dxdt', 'dydt');

  %------------------------------------------------------------------------
  %% Define some variables to make the dynamics easier to code.

  % Unpack the problem data that we use a lot into variables with short
  % names.
  s2i = schemeData.s2i;
  i2c = schemeData.i2c;
  params = schemeData.params;
  grid = schemeData.grid;
  
  %------------------------------------------------------------------------
  %% Unpack the gradient arrays.

  % Create separate arrays for each vector component in structure ds (for
  % "derivative of solution").
  for i = 1 : schemeData.vector_length
    ds.(i2c{i}) = cell(grid.dim);
    for j = 1 : grid.dim
      ds.(i2c{i}){j} = deriv{j}(schemeData.grid_index{:}, i);
    end
  end
  
  %------------------------------------------------------------------------
  %% Choose the inputs.
  for i = 1 : schemeData.vector_length
    cmpt = i2c{i};
    
    switch(params.input_choice)
      % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      case 'independent'
        
        switch(cmpt)
          case { 'x_max', 'y_max' }
            % For the upper bounds, choose the inputs to increase them.
            angular_vel_p.(cmpt) = (params.angular_vel_p(1) * (ds.(cmpt){s2i.psi} <= 0) ...
                                    + params.angular_vel_p(2) * (ds.(cmpt){s2i.psi} > 0));
            linear_acc_p.(cmpt) = (params.linear_acc_p(1) * (ds.(cmpt){s2i.vp} <= 0) ...
                                   + params.linear_acc_p(2) * (ds.(cmpt){s2i.vp} > 0));
        
          case { 'x_min', 'y_min' }
            % For the lower bounds, choose the inputs to decrease them.
            angular_vel_p.(cmpt) = (params.angular_vel_p(1) * (ds.(cmpt){s2i.psi} > 0) ...
                                    + params.angular_vel_p(2) * (ds.(cmpt){s2i.psi} <= 0));
            linear_acc_p.(cmpt) = (params.linear_acc_p(1) * (ds.(cmpt){s2i.vp} > 0) ...
                                   + params.linear_acc_p(2) * (ds.(cmpt){s2i.vp} <= 0));
        end

      % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      case 'sum'
        error('Option ''sum'' for params.input_choice is not yet implemented');

        % - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      otherwise
        error('Unknown params.input_choice: %s', params.input_choice);
    end
  end
  
  %------------------------------------------------------------------------
  %% Compute the Hamiltonian.
  
  for i = 1 : schemeData.vector_length
    cmpt = i2c{i};
    % The dynamics of the nonlinear coupled variables is the same for each
    % solution component.
    hamiltonian.(cmpt) = -(ds.(cmpt){s2i.vp} .* linear_acc_p.(cmpt) ...
                           + ds.(cmpt){s2i.psi} .* angular_vel_p.(cmpt));
    % Now include the Hamiltonian terms for the appropriate terminal
    % integrator variables.  We should subtract this term in backward time,
    % so add it in forward time.
    switch(cmpt)
      case { 'x_max', 'x_min' }
        hamiltonian.(cmpt) = hamiltonian.(cmpt) + schemeData.dxdt;
      case { 'y_max', 'y_min' }
        hamiltonian.(cmpt) = hamiltonian.(cmpt) + schemeData.dydt;
    end
  end
  
  %------------------------------------------------------------------------
  %% If necessary, restrict the update so that the reach set only grows.
  
  if schemeData.restrict_update
    for i = 1 : schemeData.vector_length
      cmpt = i2c{i};
      switch(cmpt)
        case { 'x_max', 'y_max' }
          % Restrict the update to only increase the solution.
          hamiltonian.(cmpt) = min(0, hamiltonian.(cmpt));
        case { 'x_min', 'y_min' }
          % Restrict the update to only increase the solution.
          hamiltonian.(cmpt) = max(0, hamiltonian.(cmpt));
      end
    end
  end
  
  %------------------------------------------------------------------------
  %% Pack the Hamiltonian evaluations back up into a single array.
  ham_cell = cell(schemeData.vector_length, 1);
  for i = 1 : schemeData.vector_length
    ham_cell{i} = hamiltonian.(i2c{i});
  end
  ham_value = cat(schemeData.grid.dim + 1, ham_cell{:});
  
end % function hamFunc().


%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function alpha = partialFunc(~, ~, ~, ~, schemeData, dim)
% partialFunc: Hamiltonian partial function.
%
% alpha = partialFunc(t, data, derivMin, derivMax, schemeData, dim)
%
% This function implements the partialFunc prototype for the MIE
% reachability version of the kinematic car with linear velocity and
% steering angle input.
%
% It calculates the extrema of the absolute value of the partials of the 
% analytic Hamiltonian with respect to the costate (gradient).
%
% Parameters:
%
%   t: Time at beginning of timestep.
%
%   data: Data array.
%
%   derivMin: Cell vector of minimum values of the costate (\grad \phi).
%
%   derivMax: Cell vector of maximum values of the costate (\grad \phi).
%
%   schemeData: A structure (see below).
%
%   dim: Dimension in which the partial derivatives is taken.
%
% Output Parameters:
%
%   alpha: Maximum absolute value of the partial of the Hamiltonian with
%   respect to the costate in dimension dim for the specified range of
%   costate values (O&F equation 5.12). Note that alpha can (and should) be
%   evaluated separately at each node of the grid.
%
% schemeData is a structure containing data specific to this Hamiltonian
% For this function it contains the field(s):
%
%   .params: Model parameters.
%
%   .i2s: Cell vector.  Maps from grid dimensions to variable names.

% Ian Mitchell 2011/01/26

  checkStructureFields(schemeData, 'params', 'i2s');

  % Very few terms in the Hamiltonian actually depend on the costate, so
  % this partials function is very simple to compute.
  switch(schemeData.i2s{dim})
    case 'vp'
      alpha = max(abs(schemeData.params.linear_acc_p));
    
    case 'psi'
      alpha = max(abs(schemeData.params.angular_vel_p));

    otherwise
      error('Partials do not exist in dimension %d', dim);
  end
  
end % function partialFunc().
